#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <memory.h>
#include <sys/stat.h>

int ipf;
int opf;
char *infile;
char *outfile;
struct stat ipfstat;

/***** BENCH VARIABLES ****/
struct timeval time0;
struct timeval time1;
struct timeval time2;

double realt;
double tbytes = 0;

/***** ARGUMENT VARIABLES *****/
int sinkmode; /* s */
int dbg;      /* d */
int nbuf;     /* b */
int buflen;   /* l */
int nphase;   /* p */
int nnode;    /* n */
int phase;    /* i */
int node;     /* j */
int starttime;/* t */
extern char *optarg;
extern int optind;

/***** USE BY ALL *****/
int c1,c2,c3;                   /* general counters */
int Tc;                         /* Trans. Child number */
int Rc;                         /* Rece. Child number */
char nodes[12][30];             /* Buffer for nodes.  For our case max=12 */
char *shm;                      /* pointer to the start of the shared memory in each phase */
key_t shmkey;                   /* Shared memory variables */
int shmid;
int shmsize;
int shmflg;
key_t Mshmkey;
int Mshmid;
int Mshmsize;
int Mshmflg;
int *Mcnt;
struct msemaphore *PSem;        /* IBM's msem, to get around critical area */
struct msemaphore *RSem;
struct msemaphore *TSem;
char *myhostname;
int myhostnum;
short port;
short Tport;
short Rport;
struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;
char *bufs[2];
struct hostent *addr;
unsigned long addr_tmp;
int smask;
int rtc = 0;


/***** PARENT VARIABLES *****/
int first;
int last;
int *Pcldpids;
pid_t Rcldpids[12];
pid_t Tcldpids[12];
int fromlen;
int fd;
int finished = 0;
int nfull;
int nfilled;
int nsent = 0;
int dontwait = 0;
int filllater = 0;
int stpR = 0;
int stpT = 0;
int blksize;


/***** CHILD VARIABLES *****/
int nfd;
int sTrc = 0;
int sTtc = 0;
int *RS;
int *TS;
int *TC;
msemaphore Tsemaphore;
msemaphore Rsemaphore;
msemaphore Psemaphore;

/***** PROTOTYPES *****/
void err();
void RSigToChild();
void TSigToChild();
void SigToParent1();
void SigToParent2();
void SigToMaster();
void Pchild(char *);
void Rchild(char *,char *,int,int);
void Tchild(char *,char *,int,int);
int Nread();
int Nwrite();
int pattern();
static void tvsub();

main(argc,argv)
    int argc;
    char **argv;
{
  while ((c1 = getopt(argc, argv, "s:d:l:b:p:n:j:t:")) != -1) {
    switch (c1) {
    case 's':
      sinkmode = atoi(optarg);
      break;
    case 'd':
      dbg = atoi(optarg);
      break;
    case 'l':
      buflen = atoi(optarg);
      break;
    case 'b':
      nbuf = atoi(optarg);
      break;
    case 'p':
      nphase = atoi(optarg);
      break;
    case 'n':
      nnode = atoi(optarg);
      break;
    case 'j':
      node = atoi(optarg);
      break;
    case 't':
      starttime = atol(optarg);
      break;
    default:
      goto usage;
    }
  }

  tbytes = (buflen * nbuf * nnode * nnode * (nphase-1))/1048576;
  /* Handle SIGUSR1 with function SigToMaster() */
  signal(SIGUSR1, SigToMaster);
  sigblock (sigmask(SIGUSR1));

  /* Set up shared memory for the Masters children to count and send a signal to Master */
  Mshmkey = 8576;
  Mshmsize = sizeof(int);  
  Mshmflg = 0666 | IPC_CREAT;  
  if ((Mshmid = shmget (Mshmkey, Mshmsize, Mshmflg)) == -1) err("shmget Failed!!!!");
  Mcnt = (int *) shmat(Mshmid, 0, 0);
  (*Mcnt) = 0;
  
  if ((PSem = mmap (&Psemaphore, sizeof(msemaphore), (PROT_READ | PROT_WRITE), MAP_ANONYMOUS, -1, 0)) == -1) err("Can't map Psemaphore");
  if ((msem_init(PSem, MSEM_UNLOCKED)) == NULL ) err("PSem init Failed!!!!");
  
  /* Allocate for the storage of the phase-parent pids */
  Pcldpids = malloc(nphase*sizeof(int));
  
  if (!sinkmode) {      /* if not sink mode, open and create files */
    if ((infile = getenv("INFILE")) == (char*)NULL) {
      printf("**:Environment Variable INFILE not set.\n");
      exit(1);
    }
    if ((ipf = open(infile, O_RDONLY)) < 0) {
      printf("**:Can not open infile.\n");
      exit(1);
    }
    if ((outfile=(char *)malloc(strlen(infile)+5)) == (char *)NULL) {
      printf("Couldn't allocate enough mem for outfile name\n");
      exit(1);
    }
    strcpy(outfile,infile);
    strcat(outfile,".out");
    if ((opf = creat(outfile, 0600)) < 0) {
      printf("**:Cant creat outfile.\n");
      exit(0);
    }
    stat(infile,&ipfstat);
    if (nbuf == 0) nbuf = (int) (ipfstat.st_size/buflen);
    else {
      printf("**: -n should be set to 0 if sinkmode.\n");
      exit(1);
    }
    printf("INFILE = %s; OUTFILE = %s\nipf = %d; opf = %d\n",infile,outfile,ipf,opf);
  }
  
  /* fork the children and store their pids in Pcldpids */
  for (c1=0;c1<nphase;c1++) {
    phase = c1 + 1;
    if ((*(Pcldpids + sizeof(int)*c1) = fork()) == 0) Pchild(argv[optind]);
  }

  /* wait untill the parent processes have initialized */
  while(rtc == 0) sigpause(0);
  rtc = 0;
  printf("All children ready to connect\n");
  printf("Waiting a 10 seconds to let everyone accept...\n");  /* this is plenty */
  sleep(10);
  /* send a signal to the first phase to start connecting */
  printf("\7Sending S1 to the first phase. PID: %d\n",(*(Pcldpids)));
  kill((*(Pcldpids)), SIGUSR1);

  /* wait until all the children connect */
  while(rtc == 0) sigpause(0);
  rtc = 0;
  printf("All children are connected\n");

  /* signal first phase to start transmitting */
  printf("Sleeping for 10 seconds\n");
  sleep(10);
  printf("\7Sending S1 to phase 1 to start transmitting. PID: %d\n",(*(Pcldpids)));

  gettimeofday(&time0, (struct timezone *)0);    /* get the start time */
  kill((*(Pcldpids)), SIGUSR1);    /* Tell phase parents to start  */

  while(rtc == 0) sigpause(0);     /* Wait until everyone is finished */
  gettimeofday(&time1, (struct timezone *)0);    /* get the end time */
  rtc = 0;
  printf("All finished, waiting for 5 seconds\n");
  sleep(5);

  /* Cleanup and display results */

  if (!sinkmode) {   
    close (ipf);
    close (opf);
  }
  tvsub( &time2, &time1, &time0 );
  realt = time2.tv_sec + ((double)time2.tv_usec) / 1000000;
  tbytes = (buflen/1048576) * nbuf * nnode * nnode * (nphase-1);
  printf("Transmitted bytes: %.0f MB\n",tbytes);
  printf("        Real time: %.2f Seconds\n",realt);
  printf("     Bytes/Second: %.2f MB/Seconds\n",tbytes/realt);

  printf("\n\7\7\7BYE!!!!\n");
  fflush(stdout);
  msem_remove(PSem);
  exit(0);

 usage:
  printf("WRONG USAGE!!!!!!\n");
  exit(99);
}


/*
 * main procedure for the parent phase processes 
 */
void Pchild(myhostname)
    char *myhostname;
{
  int hf;  /* host.list file identifier */
  
  /* setup signal handlers */
  signal(SIGUSR1, SigToParent1);
  signal(SIGUSR2, SigToParent2);
  smask = sigmask(SIGUSR1) | sigmask(SIGUSR2);
  sigblock (smask);

  /* Pick a phase */
  Rport = 6020 + phase;
  Tport = 6020 + phase + 1;

  /* set couple of variables for special phases */
  if (phase == 1) first = 1;
  if (phase == nphase) last = 1;

  /* size of data each child process will be sending */
  blksize = buflen / nnode;

  while ((hf = open("host.list", O_RDONLY)) < 0) {  /* open the host names file */
    if (dbg) printf("$\n");
  }

  for (c1=0;c1<nnode;c1++) {        /* find out which nodes you will connect to */
    c2 = 0;
    c3 = 0;
    while(!c2) {
      if (read(hf,&(nodes[c1][c3]),1) != 1) {
	printf("Read Error\n");
	exit(1);
      }
      if (nodes[c1][c3] == 10) c2 = 1;
      c3++;
    }
    nodes[c1][c3-1] = 0;
  }
  close(hf);

  c1 = 0;
  while(!(strcmp(myhostname, nodes[c1]))) c1++;    /* find out what your psuedo node number is */
  myhostnum = c1;

  /* setup the shared memory for use by the parent and the children */
  shmkey = 5000+(phase+1)*5-1;
  shmsize = buflen*2 + sizeof(int)*4;  
  shmflg = 0666 | IPC_CREAT;  
  if ((shmid = shmget (shmkey, shmsize, shmflg)) == -1) err("shmget Failed!!!!");
  shm = (char *) shmat(shmid, 0, 0);

  bufs[0] = shm;
  bufs[1] = shm + buflen;

  /* setup the semaphores... */
  if ((RSem = mmap (&Rsemaphore, sizeof(msemaphore), (PROT_READ | PROT_WRITE), MAP_ANONYMOUS, -1, 0)) == -1) err("Can't map Rsemaphore");
  if ((TSem = mmap (&Tsemaphore, sizeof(msemaphore), (PROT_READ | PROT_WRITE), MAP_ANONYMOUS, -1, 0)) == -1) err("Can't map Tsemaphore");
  if ((msem_init(RSem, MSEM_UNLOCKED)) == NULL ) err("RSem init Failed!!!!");
  if ((msem_init(TSem, MSEM_UNLOCKED)) == NULL ) err("TSem init Failed!!!!");

  RS = shm + buflen*2;
  TS = shm + buflen*2 + sizeof(int);
  TC = shm + buflen*2 + sizeof(int)*2;
  (*RS) = 0;
  (*TS) = 0;
  (*TC) = 0;

  if (!first) {     /* prepare and fork the receiving children */
    port = Rport;
    sinme.sin_port = htons(port);
    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) err("socket");
    if (bind(fd, &sinme, sizeof(sinme)) < 0) err("bind");
    listen(fd, 5);
    msem_lock(PSem,0);
    (*Mcnt)++;
    if (*Mcnt == nphase) {
	(*Mcnt) = 0;
	kill(getppid(),SIGUSR1);  /* I am ready to connect */
    }
    msem_unlock(PSem,0);
    for (Rc=0;Rc<nnode;Rc++) {
      if((nfd=accept(fd, &frominet, &fromlen) ) < 0) err("accept");
      if ((Rcldpids[Rc] = fork()) > 0) close (nfd);
      else if (Rcldpids[Rc] == 0) Rchild(myhostname,nodes[Rc],nbuf,blksize);
    }
    printf("%d: Receiving Children connected!\n",phase);
  }

  if (first) {
    msem_lock(PSem,0);
    (*Mcnt)++;
    if (*Mcnt == nphase) {
      (*Mcnt) = 0;
      kill(getppid(),SIGUSR1);  /* First is ready to connect */
    }
    msem_unlock(PSem,0);
    while (stpR == 0) sigpause(0);
    stpR = 0;
  }
  
  if (!last) {     /* fork the receiving children */
    for (Tc=0;Tc<nnode;Tc++) {
      if ((Tcldpids[Tc] = fork()) == 0) Tchild(myhostname,nodes[Tc],nbuf,blksize);
    }
    printf("%d: Transmitting Children connected!\n",phase);
  }

  if (last) {
    kill(getppid(),SIGUSR1);  /* Connected */
    stpR = 0;
  }

  c2 = 0;
  nfilled = 0;
  nsent = 0;

  if (first) {    /* if first phase */
    if (bufs[c2] != shm) err("Bummer!\n");
    while (stpR == 0) sigpause(0);
    stpR = 0;
    while (stpR == 0) sigpause(0); /* Getting a signal from parent means go ahead and start */
    stpR = 0;
    if(sinkmode) pattern(bufs[c2],buflen); /* if sink mode fill the buffer with the pattern func */
    else {     /* if non sink mode, fill the buffer from the file. */
      if (Nread(ipf,bufs[c2],buflen) != buflen) {
	printf("**: Couldn't read all from the disk... \n");
      }
      printf("Read first buffer\n");
    }
    while (!finished) {   
      for(c1=0;c1<nnode;c1++) {
	kill(Tcldpids[c1],SIGUSR2);  /* signal all the transmitting children to send their block */
      }
      c2 = !c2;
      if ((nsent+1) != nbuf) {      /* fill the buffer while children are sending */
	if(sinkmode) pattern(bufs[c2],buflen);
	else {
	  if (Nread(ipf,bufs[c2],buflen) != buflen) {
	    printf("**: Couldn't read all from the disk... \n");
	  }
	  printf("Read buffer %d\n",nsent);
	}
      }
      while (stpT == 0) sigpause(0);  /* wait untill children send the buffer */
      stpT = 0;
      if ((nsent) == nbuf) finished = 1;
    }
  }
  else if (last) {
    c1 = 0;
    while (!finished) {
      for(c1=0;c1<nnode;c1++){  /* signal all the receiving children to read the buffer */
	kill(Rcldpids[c1],SIGUSR1);
      }
      while (stpR == 0) sigpause(0);  /* wait untill the buffer is read */
      stpR = 0;
      if (!sinkmode) {    /* if non sink mode, write the buffer to a file. */
	if (Nwrite(opf,bufs[c2],buflen) != buflen) {
	  printf("**: Couldn't write all for the buffer to the disk.\n");
	}
      }
      if ((nfilled) == nbuf) finished = 1;
    }
  }
  else {
    while (stpR == 0) sigpause(0);
    stpR = 0;
    nfull = 0;
    nfilled = 0;
    stpR = 0;
    stpT = 0;
    dontwait = 0;
    for(c1=0;c1<nnode;c1++) {
      kill(Rcldpids[c1],SIGUSR1);
    }
    while (stpR == 0) sigpause(0);
    stpR = 0;

    while (!finished) {
      /* signal receivers to receive */
      if (nfilled != nbuf) for(c1=0;c1<nnode;c1++) kill(Rcldpids[c1],SIGUSR1);
      else dontwait = 1;
      /* signal transmiters to send */
      if (nsent != nbuf) for(c1=0;c1<nnode;c1++) kill(Tcldpids[c1],SIGUSR2);
      else {
	finished = 1;
	continue;
      }

      if (!dontwait) {
	while (stpR == 0) sigpause(0);  /* wait until the buffer is filled unless, */
	stpR = 0;                       /* the last one was already to filled */
      }
      while (stpT == 0) sigpause(0);    /* wait untill the buffer is sent */
      stpT = 0;
    }
  }
  /***** CLEANUP *****/
  msem_remove(RSem);
  msem_remove(TSem);
  msem_lock(PSem,0);
  (*Mcnt)++;
  if (*Mcnt == nphase) { 
    (*Mcnt) = 0;
    kill(getppid(),SIGUSR1);
  }
  msem_unlock(PSem,0);
  fflush(stdout);
  exit(0);
}



void SigToParent1() {    /* Signal handler to parent from the receiving children */
  stpR++;
  nfull++;
  nfilled++;
}

void SigToParent2() {    /* Signal handler to parent from the sending children */
  stpT++;
  nfull--;
  nsent++;
}

void SigToMaster() {     /* Signal to the master process from the parent phase processes */
  rtc++;
}

void RSigToChild() {     /* Signal to the receiving children from the parent to fill the buffers */
  sTrc = 1;
}

void TSigToChild() {     /* Signal to the transmitting children from the parent to send the buffer */
  sTtc = 1;
}

void err(s)
    char *s;
{
  perror(s);
  exit(1);
}




void Rchild(me, cwith, nbuf, blksize)
    char *me;
    char *cwith;
    int nbuf;
    int blksize;
{
  close (fd);
  c2 = 0;
  signal(SIGUSR1, RSigToChild);
  sigblock (sigmask(SIGUSR1));
  for (c1=0;c1<nbuf;c1++) {             /* loop untill nbufs are received */
    while (sTrc == 0) sigpause(0);      /* read after receiving the appropriate signal */
    sTrc = 0;
    if ((Nread(nfd,bufs[c2]+blksize*Rc,blksize)) != blksize){
      printf("%d-%dR: Couldn't receive all of the block\n",phase,Rc);
      exit(1);
    }
    c2 = !c2;
    msem_lock(RSem,0);                  /* take care of the race condition. */
    if ((++(*RS)) == nnode) {           /* if this child is the last one to read, send    */
      (*RS) = 0;                        /* a signal to parent, saying "filled the buffer" */
      kill(getppid(),SIGUSR1);
    }
    msem_unlock(RSem,0);
  }
  close (nfd);
  exit(0);
}


void Tchild(me, cwith, nbuf, blksize)
    char *me;
    char *cwith;
    int nbuf;
    int blksize;
{
  signal(SIGUSR2, TSigToChild);
  sigblock (sigmask(SIGUSR2));
  port = Tport;
  if ((addr=gethostbyname(cwith)) == NULL) err("bad host name");
  sinhim.sin_family = addr->h_addrtype;
  bcopy(addr->h_addr,(char*)&addr_tmp, addr->h_length);
#if defined(cray)
  sinhim.sin_addr = addr_tmp;
#else
  sinhim.sin_addr.s_addr = addr_tmp;
#endif /* cray */
  sinhim.sin_port =  htons(port);
  sinme.sin_port = 0;		/* free choice */

  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      err("socket");
  if ((connect(fd, &sinhim, sizeof(sinhim) )) < 0) err("can't connect");
  printf("%d-%dT: Connected!\n",phase,Tc);
  msem_lock(TSem,0);                   /* take care of race condition */
  if ((++(*TC)) == nnode) {            /* if I am the last one to connect, */
    kill(getppid(),SIGUSR1);           /* send a signal to parent signifying ready */
    *TC= 0;
  }
  msem_unlock(TSem,0);
  
  for (c1=0;c1<nbuf;c1++) {            /* loop untill all nbuf buffers are sent */
    while (sTtc == 0) sigpause(0);     /* wait for a send signal */
    sTtc = 0;
    if ((Nwrite(fd,bufs[c2]+blksize*Tc,blksize)) != blksize) {
      printf("%d-%dT: Couldn't transfer all",phase,Tc);
      exit(1);
    }
    c2 = !c2;
    msem_lock(TSem,0);                 /* take care of race condition */
    if ((++(*TS)) == nnode) {          /* if this child is the last one to send a block, */
      kill(getppid(),SIGUSR2);         /* signal parent saying buffer is sent*/
      (*TS) = 0;
    }
    msem_unlock(TSem,0);
  }
  close (fd);
  fflush(stdout);
  exit(0);
}
pattern( cp, cnt )                     /* to fill the sync mode buffers */
     register char *cp;
     register int cnt;
{
  register char c;
  c = 0;
	while( cnt-- > 0 )  {
	  while( !isprint((c&0x7F)) )  c++;
	  *cp++ = (c++&0x7F);
	}
}

/*
 *			N R E A D
 */
int Nread( fd, buf, count )
     int fd;
     void *buf;
     int count;
{
  register unsigned cnt = 0;
  register int nread;

  do {
    nread = read(fd,buf,count-cnt);
    if(nread < 0)  {
      perror("apt_mread");
      return(-1);
    }
    if(nread == 0) return ((int)cnt);
    cnt += (unsigned)nread;
    buf += nread;
  } while (cnt < count);
  return((int)cnt);
}

/*
 *			N W R I T E
 */
Nwrite( fd, buf, count )
     int fd;
     void *buf;
     int count;
{
  register int cnt;
  cnt = write( fd, buf, count );
  return(cnt);
}

/*
 *  tvsub()
 */

static void
tvsub(tdiff, t1, t0)
     struct timeval *tdiff, *t1, *t0;
{
  
  tdiff->tv_sec = t1->tv_sec - t0->tv_sec;
  tdiff->tv_usec = t1->tv_usec - t0->tv_usec;
  if (tdiff->tv_usec < 0) tdiff->tv_sec--, tdiff->tv_usec += 1000000;
}
